"""Backend mixins for Ralph."""

import json
import logging

from ralph.conf import settings

logger = logging.getLogger(__name__)


class HistoryMixin:
    """Backend history mixin.

    Handle backend download history to avoid fetching same files multiple
    times if they are already available.
    """

    @property
    def history(self):
        """Get backend history."""
        logger.debug("Loading history file: %s", str(settings.HISTORY_FILE))

        if not hasattr(self, "_history"):
            try:
                with settings.HISTORY_FILE.open(
                    encoding=settings.LOCALE_ENCODING
                ) as history_file:
                    self._history = json.load(history_file)
            except FileNotFoundError:
                self._history = []
        return self._history

    def write_history(self, history):
        """Write given history as a JSON file."""
        logger.debug("Writing history file: %s", str(settings.HISTORY_FILE))

        if not settings.HISTORY_FILE.parent.exists():
            settings.HISTORY_FILE.parent.mkdir(parents=True)

        with settings.HISTORY_FILE.open(
            "w", encoding=settings.LOCALE_ENCODING
        ) as history_file:
            json.dump(history, history_file)

        # Update history
        self._history = history

    def clean_history(self, selector):
        """Clean selected events from the history.

        selector: a callable that selects events that need to be removed
        """
        self._history = list(filter(lambda event: not selector(event), self.history))
        self.write_history(self._history)

    def append_to_history(self, event):
        """Append event to history."""
        self.write_history(self.history + [event])

    def get_command_history(self, backend_name, command):
        """Extract entry ids from the history for a given command and backend_name."""

        def filter_by_name_and_command(entry):
            """Check whether the history entry matches the backend_name and command."""
            return entry.get("backend") == backend_name and (
                command in [entry.get("command"), entry.get("action")]
            )

        return [
            entry["id"] for entry in filter(filter_by_name_and_command, self.history)
        ]



"""Virtual classroom xAPI events result fields definitions."""

from pydantic import StrictStr

from ..base.results import BaseXapiResult


class VirtualClassroomAnsweredPollResult(BaseXapiResult):
    """Pydantic model for virtual classroom answered poll `result` property.

    Attributes:
        response (str): Consists of the response for the given Activity.
    """

    response: StrictStr  # = StrictStr()



"""Tests for Ralph's configuration loading."""

from importlib import reload

import pytest

from ralph import conf
from ralph.backends.data.es import ESDataBackend
from ralph.conf import (
    AuthBackend,
    AuthBackends,
    CommaSeparatedTuple,
    Settings,
    settings,
)
from ralph.exceptions import ConfigurationException


def test_conf_settings_field_value_priority(fs, monkeypatch):
    """Test that the Settings object field values are defined in the following
    descending order of priority:

        1. Arguments passed to the initializer.
        2. Environment variables.
        3. Dotenv variables (.env)
        4. Default values.
    """

    # 4. Using default value.
    assert str(Settings().AUTH_FILE) == str(settings.APP_DIR / "auth.json")

    # 3. Using dotenv variables (overrides default value).
    fs.create_file(".env", contents="RALPH_AUTH_FILE=/dotenv\n")
    assert str(Settings().AUTH_FILE) == "/dotenv"

    # 2. Using environment variable value (overrides dotenv value).
    monkeypatch.setenv("RALPH_AUTH_FILE", "/environment")
    assert str(Settings().AUTH_FILE) == "/environment"

    # 1. Using argument value (overrides environment value).
    assert str(Settings(AUTH_FILE="/argument").AUTH_FILE) == "/argument"


@pytest.mark.parametrize(
    "value,expected",
    [
        ("foo", ("foo",)),
        (("foo",), ("foo",)),
        (["foo"], ("foo",)),
        ("foo,bar,baz", ("foo", "bar", "baz")),
    ],
)
def test_conf_comma_separated_list_with_valid_values(value, expected, monkeypatch):
    """Test the CommaSeparatedTuple pydantic data type with valid values."""
    assert next(CommaSeparatedTuple.__get_validators__())(value) == expected
    monkeypatch.setenv("RALPH_BACKENDS__DATA__ES__HOSTS", "".join(value))
    assert ESDataBackend().settings.HOSTS == expected


@pytest.mark.parametrize("value", [{}, None])
def test_conf_comma_separated_list_with_invalid_values(value):
    """Test the CommaSeparatedTuple pydantic data type with invalid values."""
    with pytest.raises(TypeError, match="Invalid comma-separated list"):
        next(CommaSeparatedTuple.__get_validators__())(value)


@pytest.mark.parametrize(
    "value,is_valid,expected",
    [
        ("oidc", True, (AuthBackend.OIDC,)),
        ("basic", True, (AuthBackend.BASIC,)),
        ("bASIc", True, (AuthBackend.BASIC,)),
        ("oidc,basic", True, (AuthBackend.OIDC, AuthBackend.BASIC)),
        ("notvalid", False, None),
        ("basic,notvalid", False, None),
    ],
)
def test_conf_auth_backend(value, is_valid, expected, monkeypatch):
    """Test the AuthBackends data type with valid and invalid values."""
    if is_valid:
        assert next(AuthBackends.__get_validators__())(value) == expected
        monkeypatch.setenv("RALPH_RUNSERVER_AUTH_BACKENDS", "".join(value))
        reload(conf)
        assert conf.settings.RUNSERVER_AUTH_BACKENDS == expected
    else:
        with pytest.raises(ValueError, match="'notvalid' is not a valid AuthBackend"):
            next(AuthBackends.__get_validators__())(value)


def test_conf_core_settings_should_impact_settings_defaults(monkeypatch):
    """Test that core settings update application settings values."""
    monkeypatch.setenv("RALPH_APP_DIR", "/foo")
    monkeypatch.setenv("RALPH_LOCALE_ENCODING", "ascii")
    reload(conf)

    # Configuration.
    assert conf.Settings.Config.env_file_encoding == "ascii"

    # Properties.
    assert str(conf.settings.APP_DIR) == "/foo"
    assert conf.settings.LOCALE_ENCODING == "ascii"

    # Defaults.
    assert str(conf.settings.AUTH_FILE) == "/foo/auth.json"


def test_conf_forbidden_scopes_without_authority(monkeypatch):
    """Test that using RESTRICT_BY_SCOPES without RESTRICT_BY_AUTHORITY raises an
    error."""

    monkeypatch.setenv("RALPH_LRS_RESTRICT_BY_AUTHORITY", False)
    monkeypatch.setenv("RALPH_LRS_RESTRICT_BY_SCOPES", True)

    with pytest.raises(
        ConfigurationException,
        match=(
            "LRS_RESTRICT_BY_AUTHORITY must be set to True if using "
            "LRS_RESTRICT_BY_SCOPES=True"
        ),
    ):
        reload(conf)



"""Base xAPI `Object` definitions (1)."""

import sys
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import AnyUrl, StrictStr, constr, validator

from ..config import BaseModelWithConfig
from .common import IRI, LanguageMap

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseXapiActivityDefinition(BaseModelWithConfig):
    """Pydantic model for `Activity` type `definition` property.

    Attributes:
        name (LanguageMap): Consists of the human-readable/visual name of the Activity.
        description (LanguageMap): Consists of a description of the Activity.
        type (IRI): Consists of the type of the Activity.
        moreInfo (URL): Consists of an URL to a document about the Activity.
        extensions (dict): Consists of a dictionary of other properties as needed.
    """

    name: Optional[LanguageMap]
    description: Optional[LanguageMap]
    type: Optional[IRI]
    moreInfo: Optional[AnyUrl]
    extensions: Optional[Dict[IRI, Union[str, int, bool, list, dict, None]]]


class BaseXapiInteractionComponent(BaseModelWithConfig):
    """Pydantic model for an interaction component.

    Attributes:
        id (str): Consists of an identifier of the interaction component.
        description (LanguageMap): Consists of the description of the interaction.
    """

    id: constr(regex=r"^[^\s]+$")
    description: Optional[LanguageMap]


class BaseXapiActivityInteractionDefinition(BaseXapiActivityDefinition):
    """Pydantic model for `Activity` type `definition` property.

    It is defined for field with interaction properties.

    Attributes:
        interactionType (str): Consists of the type of the interaction.
        correctResponsesPattern (list): Consists of a pattern for the correct response.
        choices (list): Consists of a list of selectable choices.
        scale (list): Consists of a list of the options on the `likert` scale.
        source (list): Consists of a list of sources to be matched.
        target (list): Consists of a list of targets to be matched.
        steps (list): Consists of a list of the elements making up the interaction.
    """

    interactionType: Literal[
        "true-false",
        "choice",
        "fill-in",
        "long-fill-in",
        "matching",
        "performance",
        "sequencing",
        "likert",
        "numeric",
        "other",
    ]
    correctResponsesPattern: Optional[List[StrictStr]]
    choices: Optional[List[BaseXapiInteractionComponent]]
    scale: Optional[List[BaseXapiInteractionComponent]]
    source: Optional[List[BaseXapiInteractionComponent]]
    target: Optional[List[BaseXapiInteractionComponent]]
    steps: Optional[List[BaseXapiInteractionComponent]]

    @validator("choices", "scale", "source", "target", "steps")
    @classmethod
    def check_unique_ids(cls, value: Any) -> None:
        """Check the uniqueness of interaction components IDs."""
        if len(value) != len({x.id for x in value}):
            raise ValueError("Duplicate InteractionComponents are not valid")


class BaseXapiActivity(BaseModelWithConfig):
    """Pydantic model for `Activity` type property.

    Attributes:
        id (IRI): Consists of an identifier for a single unique Activity.
        objectType (str): Consists of the value `Activity`.
        definition (dict): See BaseXapiActivityDefinition and
            BaseXapiActivityInteractionDefinition.
    """

    id: IRI
    objectType: Optional[Literal["Activity"]]
    definition: Optional[
        Union[
            BaseXapiActivityDefinition,
            BaseXapiActivityInteractionDefinition,
        ]
    ]


class BaseXapiStatementRef(BaseModelWithConfig):
    """Pydantic model for `StatementRef` type property.

    Attributes:
        objectType (str): Consists of the value `StatementRef`.
        id (UUID): Consists of the UUID of the referenced statement.
    """

    id: UUID
    objectType: Literal["StatementRef"]


BaseXapiUnnestedObject = Union[BaseXapiActivity, BaseXapiStatementRef]



"""Basic authentication & authorization related tools for the Ralph API."""

import logging
from functools import lru_cache
from pathlib import Path
from threading import Lock
from typing import Any, Iterator, List, Optional

import bcrypt
from cachetools import TTLCache, cached
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from pydantic import BaseModel, root_validator
from starlette.authentication import AuthenticationError

from ralph.api.auth.user import AuthenticatedUser
from ralph.conf import settings

# Unused password used to avoid timing attacks, by comparing passwords supplied
# with invalid credentials to something innocuous with the same method as if
# it were a legitimate user.
UNUSED_PASSWORD = bcrypt.hashpw(b"ralph", bcrypt.gensalt())


security = HTTPBasic(auto_error=False)

# API auth logger
logger = logging.getLogger(__name__)


class UserCredentials(AuthenticatedUser):
    """Pydantic model for user credentials as stored in the credentials file.

    Attributes:
        hash (str): Consists of the hashed password for a declared user.
        username (str): Consists of the username for a declared user.
    """

    hash: str
    username: str


class ServerUsersCredentials(BaseModel):
    """Custom root pydantic model.

    Describe expected list of all server users credentials as stored in
    the credentials file.

    Attributes:
        __root__ (List): Custom root consisting of the
                        list of all server users credentials.
    """

    __root__: List[UserCredentials]

    def __add__(self, other) -> Any:  # noqa: D105
        return ServerUsersCredentials.parse_obj(self.__root__ + other.__root__)

    def __getitem__(self, item: int) -> UserCredentials:  # noqa: D105
        return self.__root__[item]

    def __len__(self) -> int:  # noqa: D105
        return len(self.__root__)

    def __iter__(self) -> Iterator[UserCredentials]:  # noqa: D105
        return iter(self.__root__)

    @root_validator
    @classmethod
    def ensure_unique_username(cls, values: Any) -> Any:
        """Every username should be unique among registered users."""
        usernames = [entry.username for entry in values.get("__root__")]
        if len(usernames) != len(set(usernames)):
            raise ValueError(
                "You cannot create multiple credentials with the same username"
            )
        return values


@lru_cache()
def get_stored_credentials(auth_file: Path) -> ServerUsersCredentials:
    """Helper to read the credentials/scopes file.

    Read credentials from JSON file and stored them to avoid reloading them with every
    request.

    Args:
        auth_file (Path): Path to the JSON credentials scope file.

    Returns:
        credentials (ServerUsersCredentials): Cache-memorized credentials.

    """
    auth_file = Path(auth_file)
    if not auth_file.exists():
        msg = "Credentials file <%s> not found."
        logger.warning(msg, auth_file)
        raise AuthenticationError(msg.format(auth_file))
    return ServerUsersCredentials.parse_file(auth_file)


@cached(
    TTLCache(maxsize=settings.AUTH_CACHE_MAX_SIZE, ttl=settings.AUTH_CACHE_TTL),
    lock=Lock(),
    key=lambda credentials: (
        credentials.username,
        credentials.password,
    )
    if credentials is not None
    else None,
)
def get_basic_auth_user(
    credentials: Optional[HTTPBasicCredentials] = Depends(security),
) -> AuthenticatedUser:
    """Check valid auth parameters.

    Get the basic auth parameters from the Authorization header, and checks them
    against our own list of hashed credentials.

    Args:
        credentials (iterator): auth parameters from the Authorization header

    Raises:
        HTTPException
    """
    if not credentials:
        logger.debug("No credentials were found for Basic auth")
        return None

    try:
        user = next(
            filter(
                lambda u: u.username == credentials.username,
                get_stored_credentials(settings.AUTH_FILE),
            )
        )
        hashed_password = user.hash
    except StopIteration:
        # next() gets the first item in the enumerable; if there is none, it
        # raises a StopIteration error as it is out of bounds.
        logger.warning(
            "User %s tried to authenticate but this account does not exists",
            credentials.username,
        )
        hashed_password = None
    except AuthenticationError as exc:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN, detail=str(exc)
        ) from exc

    # Check that a password was passed
    if not hashed_password:
        # We're doing a bogus password check anyway to avoid timing attacks on
        # usernames
        bcrypt.checkpw(
            credentials.password.encode(settings.LOCALE_ENCODING), UNUSED_PASSWORD
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid authentication credentials",
            headers={"WWW-Authenticate": "Basic"},
        )

    # Check password validity
    if not bcrypt.checkpw(
        credentials.password.encode(settings.LOCALE_ENCODING),
        hashed_password.encode(settings.LOCALE_ENCODING),
    ):
        logger.warning(
            "Authentication failed for user %s",
            credentials.username,
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid authentication credentials",
            headers={"WWW-Authenticate": "Basic"},
        )

    user = AuthenticatedUser(
        scopes=user.scopes,
        agent=dict(user.agent),
        target=user.target,
    )

    return user



"""`Audio` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Audio


class AudioActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for audio `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `https://w3id.org/xapi/audio/activity-type/audio`.
    """

    type: Literal["https://w3id.org/xapi/audio/activity-type/audio"]


class AudioActivity(BaseXapiActivity):
    """Pydantic model for audio `Activity` type.

    Attributes:
        name (dict): Consists of the dictionary `{"en-US": <name of the audio>}`.
        definition (dict): See audioActivityDefinition.
    """

    definition: AudioActivityDefinition



"""Base xAPI `Result` definitions."""

from datetime import timedelta
from decimal import Decimal
from typing import Any, Dict, Optional, Union

from pydantic import StrictBool, StrictStr, conint, root_validator

from ..config import BaseModelWithConfig
from .common import IRI


class BaseXapiResultScore(BaseModelWithConfig):
    """Pydantic model for result `score` property.

    Attributes:
        scaled (int): Consists of the normalized score related to the experience.
        raw (Decimal): Consists of the non-normalized score achieved by the Actor.
        min (Decimal): Consists of the lowest possible score.
        max (Decimal): Consists of the highest possible score.
    """

    scaled: Optional[conint(ge=-1, le=1)]
    raw: Optional[Decimal]
    min: Optional[Decimal]
    max: Optional[Decimal]

    @root_validator
    @classmethod
    def check_raw_min_max_relation(cls, values: Any) -> Any:
        """Check the relationship `min < raw < max`."""
        raw_value = values.get("raw", None)
        min_value = values.get("min", None)
        max_value = values.get("max", None)

        if min_value:
            if max_value and min_value > max_value:
                raise ValueError("min cannot be greater than max")
            if raw_value and min_value > raw_value:
                raise ValueError("min cannot be greater than raw")
        if max_value:
            if raw_value and raw_value > max_value:
                raise ValueError("raw cannot be greater than max")

        return values


class BaseXapiResult(BaseModelWithConfig):
    """Pydantic model for `result` property.

    Attributes:
        score (dict): See BaseXapiResultScore.
        success (bool): Indicates whether the attempt on the Activity was successful.
        completion (bool): Indicates whether the Activity was completed.
        response (str): Consists of the response for the given Activity.
        duration (timedelta): Consists of the duration over which the Statement
            occurred.
        extensions (dict): Consists of a dictionary of other properties as needed.
    """

    score: Optional[BaseXapiResultScore]
    success: Optional[StrictBool]
    completion: Optional[StrictBool]
    response: Optional[StrictStr]
    duration: Optional[timedelta]
    extensions: Optional[Dict[IRI, Union[str, int, bool, list, dict, None]]]



"""Event validator definition."""


import json
import logging
from typing import Any, Generator, Optional, TextIO

from pydantic import ValidationError

from ralph.exceptions import BadFormatException, UnknownEventException
from ralph.models.selector import ModelSelector

logger = logging.getLogger(__name__)


class Validator:
    """Events validator using pydantic models."""

    def __init__(self, model_selector: ModelSelector):
        """Initialize Validator."""
        self.model_selector = model_selector

    def validate(
        self, input_file: TextIO, ignore_errors: bool, fail_on_unknown: bool
    ) -> Generator:
        """Validate JSON event strings line by line."""
        total = 0
        success = 0
        for event_str in input_file:
            try:
                total += 1
                yield self._validate_event(event_str)
                success += 1
            except (json.JSONDecodeError, TypeError) as err:
                message = "Input event is not a valid JSON string"
                self._log_error(message, event_str, err)
                if not ignore_errors:
                    raise BadFormatException(message) from err
            except UnknownEventException as err:
                self._log_error(err, event_str)
                if fail_on_unknown:
                    raise err
            except ValidationError as err:
                message = f"Input event is not a valid {err.model.__name__} event."
                self._log_error(message, event_str, err)
                if not ignore_errors:
                    raise BadFormatException(message) from err
        logger.info("Total events: %d, Invalid events: %d", total, total - success)

    def get_first_valid_model(self, event: dict) -> Any:
        """Return the first successfully instantiated model for the event.

        Raises:
            UnknownEventException: When the event does not match any model.
            ValidationError: When the last validated event is invalid.
        """
        error: Optional[BaseException] = None
        for model in self.model_selector.get_models(event):
            try:
                return model(**event)
            except ValidationError as err:
                error = err

        raise error

    def _validate_event(self, event_str: str) -> Any:
        """Validate a single JSON string event.

        Raises:
            TypeError: When the event_str is not of type string.
            JSONDecodeError: When the event_str is not a valid JSON string.
            UnknownEventException: When no matching model is found for the event.
            ValidationError: When the event is failing the pydantic model validation.

        Returns:
            event_str (str): The cleaned JSON-formatted input event_str.
        """
        event = json.loads(event_str)
        return self.get_first_valid_model(event).json()

    @staticmethod
    def _log_error(
        message: object, event_str: str, error: Optional[BaseException] = None
    ) -> None:
        logger.error(message)
        logger.debug("Raised error: %s, for event : %s", error, event_str)



"""xAPI pydantic models."""

# flake8: noqa

from .lms.statements import (
    LMSAccessedFile,
    LMSAccessedPage,
    LMSDownloadedAudio,
    LMSDownloadedDocument,
    LMSDownloadedFile,
    LMSDownloadedVideo,
    LMSRegisteredCourse,
    LMSUnregisteredCourse,
    LMSUploadedAudio,
    LMSUploadedDocument,
    LMSUploadedFile,
    LMSUploadedVideo,
)
from .navigation.statements import PageTerminated, PageViewed
from .video.statements import (
    VideoCompleted,
    VideoEnableClosedCaptioning,
    VideoInitialized,
    VideoPaused,
    VideoPlayed,
    VideoScreenChangeInteraction,
    VideoSeeked,
    VideoTerminated,
    VideoVolumeChangeInteraction,
)
from .virtual_classroom.statements import (
    VirtualClassroomAnsweredPoll,
    VirtualClassroomInitialized,
    VirtualClassroomJoined,
    VirtualClassroomLeft,
    VirtualClassroomLoweredHand,
    VirtualClassroomMuted,
    VirtualClassroomPostedPublicMessage,
    VirtualClassroomRaisedHand,
    VirtualClassroomSharedScreen,
    VirtualClassroomStartedCamera,
    VirtualClassroomStartedPoll,
    VirtualClassroomStoppedCamera,
    VirtualClassroomTerminated,
    VirtualClassroomUnmuted,
    VirtualClassroomUnsharedScreen,
)



"""Model selector definition."""

from collections import Counter
from dataclasses import dataclass
from importlib import import_module
from inspect import getmembers, isclass
from itertools import chain
from types import ModuleType
from typing import Any, Dict, List, Tuple, Union

from pydantic import BaseModel

from ralph.conf import MODEL_PATH_SEPARATOR
from ralph.exceptions import UnknownEventException
from ralph.utils import get_dict_value_from_path


@dataclass(frozen=True)
class LazyModelField:
    """Model field."""

    path: Tuple[str]

    def __init__(self, path: str) -> None:
        """Initialize Lazy Model Field."""
        object.__setattr__(self, "path", tuple(path.split(MODEL_PATH_SEPARATOR)))


@dataclass(frozen=True)
class Rule:
    """Rule used for selection."""

    field: LazyModelField
    value: Union[LazyModelField, Any]

    def check(self, event: Dict) -> bool:
        """Check if event matches the rule.

        Args:
            event (dict): The event to check.
        """
        event_value = get_dict_value_from_path(event, self.field.path)
        expected_value = self.value
        if isinstance(expected_value, LazyModelField):
            expected_value = get_dict_value_from_path(event, expected_value.path)
        return event_value == expected_value


def selector(**filters: Any) -> List[Rule]:
    """Return a list of rules that should match in order to select an event.

    Args:
        **filters: Each keyword represents the path to a field, the corresponding
            value represents the value which the field should match.
            Example: `event_type="server"` will generate a Rule verifying that
            `event.event_type == "server"`
    """
    return [Rule(LazyModelField(field), value) for field, value in filters.items()]


class ModelSelector:
    """Matching model selector for a given event.

    Attributes:
        model_rules (dict): Stores the list of rules for each model.
        decision_tree (dict): Stores the rule checking order for model selection.
    """

    def __init__(self, module: str = "ralph.models.edx") -> None:
        """Instantiate ModelSelector."""
        self.model_rules = ModelSelector.build_model_rules(import_module(module))
        self.decision_tree = self.get_decision_tree(self.model_rules)

    @staticmethod
    def build_model_rules(module: ModuleType) -> Dict:
        """Build the model_rules dictionary.

        Using BaseModel classes defined in the module.
        """
        model_rules = {}
        for _, class_ in getmembers(module, isclass):
            if issubclass(class_, BaseModel) and hasattr(class_, "__selector__"):
                model_rules[class_] = class_.__selector__
        return model_rules

    def get_first_model(self, event: Dict) -> Any:
        """Return the first matching model for the event. See `self.get_models`."""
        return self.get_models(event)[0]

    def get_models(self, event: dict, tree=None):
        """Recursively go through the decision tree to find the event matching models.

        Args:
            event (dict): Event to retrieve the corresponding model.
            tree (dict): The (sub) decision tree, `None` stands for the whole decision
                tree.

        Returns:
            models (list of BaseModels): When the event matches all rules of the models.

        Raises:
            UnknownEventException: When the event does not match any model.
        """
        if tree is None:
            tree = self.decision_tree
        rule = next(iter(tree))
        is_valid = rule.check(event)
        subtree = tree[rule][is_valid]
        if not isinstance(subtree, dict):
            if subtree is None:
                raise UnknownEventException(
                    "No matching pydantic model found for input event"
                )
            # Here we have found the model.
            return subtree
        return self.get_models(event, subtree)

    def get_decision_tree(self, model_rules):
        """Recursively construct the decision tree."""
        rule_counter = Counter(chain.from_iterable(model_rules.values()))
        if not rule_counter:
            return list(model_rules)

        # We retrieve the rule with the highest occurrence and use it to split
        # the decision tree in two subtrees:
        # 1 - true_subtree: a subtree of models that require the most occurred
        #     rule to be true.
        # 2 - false_subtree: a subtree of models that don't use the most
        #     occurred rule.
        root_rule = rule_counter.most_common(1)[0][0]
        true_subtree = {}
        false_subtree = {}
        for model, rules in model_rules.items():
            if root_rule not in rules:
                false_subtree[model] = rules
                continue
            true_subtree[model] = list(filter(lambda rule: rule != root_rule, rules))

        if sorted(true_subtree.values(), key=lambda x: -len(x))[0]:
            true_subtree = self.get_decision_tree(true_subtree)
        else:
            true_subtree = list(true_subtree)
        false_subtree = self.get_decision_tree(false_subtree) if false_subtree else None

        return {root_rule: {True: true_subtree, False: false_subtree}}



"""MongoDB LRS backend for Ralph."""

import logging
from typing import Iterator, List, Optional

from bson.objectid import ObjectId
from pymongo import ASCENDING, DESCENDING

from ralph.backends.data.mongo import (
    MongoDataBackend,
    MongoDataBackendSettings,
    MongoQuery,
)
from ralph.backends.lrs.base import (
    AgentParameters,
    BaseLRSBackend,
    BaseLRSBackendSettings,
    RalphStatementsQuery,
    StatementQueryResult,
)
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException

logger = logging.getLogger(__name__)


class MongoLRSBackendSettings(BaseLRSBackendSettings, MongoDataBackendSettings):
    """MongoDB LRS backend default configuration."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__LRS__MONGO__"


class MongoLRSBackend(BaseLRSBackend[MongoLRSBackendSettings], MongoDataBackend):
    """MongoDB LRS backend."""

    def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the results of a statements query using xAPI parameters."""
        query = self.get_query(params)
        try:
            mongo_response = list(
                self.read(query=query, target=target, chunk_size=params.limit)
            )
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from MongoDB")
            raise error

        search_after = None
        if mongo_response:
            search_after = mongo_response[-1]["_id"]

        return StatementQueryResult(
            statements=[document["_source"] for document in mongo_response],
            pit_id=None,
            search_after=search_after,
        )

    def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> Iterator[dict]:
        """Yield statements with matching ids from the backend."""
        query = self.query_class(filter={"_source.id": {"$in": ids}})
        try:
            mongo_response = self.read(query=query, target=target)
            yield from (document["_source"] for document in mongo_response)
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from MongoDB")
            raise error

    @staticmethod
    def get_query(params: RalphStatementsQuery) -> MongoQuery:
        """Construct query from statement parameters."""
        mongo_query_filters = {}

        if params.statement_id:
            mongo_query_filters.update({"_source.id": params.statement_id})

        MongoLRSBackend._add_agent_filters(mongo_query_filters, params.agent, "actor")
        MongoLRSBackend._add_agent_filters(
            mongo_query_filters, params.authority, "authority"
        )

        if params.verb:
            mongo_query_filters.update({"_source.verb.id": params.verb})

        if params.activity:
            mongo_query_filters.update(
                {
                    "_source.object.objectType": "Activity",
                    "_source.object.id": params.activity,
                },
            )

        if params.since:
            mongo_query_filters.update({"_source.timestamp": {"$gt": params.since}})

        if params.until:
            if not params.since:
                mongo_query_filters["_source.timestamp"] = {}
            mongo_query_filters["_source.timestamp"].update({"$lte": params.until})

        if params.search_after:
            search_order = "$gt" if params.ascending else "$lt"
            mongo_query_filters.update(
                {"_id": {search_order: ObjectId(params.search_after)}}
            )

        mongo_sort_order = ASCENDING if params.ascending else DESCENDING
        mongo_query_sort = [
            ("_source.timestamp", mongo_sort_order),
            ("_id", mongo_sort_order),
        ]

        # Note: `params` fields are validated thus we skip MongoQuery validation.
        return MongoQuery.construct(
            filter=mongo_query_filters, limit=params.limit, sort=mongo_query_sort
        )

    @staticmethod
    def _add_agent_filters(
        mongo_query_filters: dict, agent_params: AgentParameters, target_field: str
    ) -> None:
        """Add filters relative to agents to mongo_query_filters.

        Args:
            mongo_query_filters (dict): Filters passed to MongoDB query.
            agent_params (AgentParameters): Agent query parameters to search for.
            target_field (str): The target agent field name to perform the search.
        """
        if not agent_params:
            return

        if not isinstance(agent_params, dict):
            agent_params = agent_params.dict()

        if agent_params.get("mbox"):
            key = f"_source.{target_field}.mbox"
            mongo_query_filters.update({key: agent_params.get("mbox")})

        if agent_params.get("mbox_sha1sum"):
            key = f"_source.{target_field}.mbox_sha1sum"
            mongo_query_filters.update({key: agent_params.get("mbox_sha1sum")})

        if agent_params.get("openid"):
            key = f"_source.{target_field}.openid"
            mongo_query_filters.update({key: agent_params.get("openid")})

        if agent_params.get("account__name"):
            key = f"_source.{target_field}.account.name"
            mongo_query_filters.update({key: agent_params.get("account__name")})
            key = f"_source.{target_field}.account.homePage"
            mongo_query_filters.update({key: agent_params.get("account__home_page")})



"""Video event model definitions."""

import sys
from typing import Optional, Union

from pydantic import Json

from ralph.models.edx.video.fields.events import (
    PauseVideoEventField,
    PlayVideoEventField,
    SeekVideoEventField,
    SpeedChangeVideoEventField,
    StopVideoEventField,
    VideoBaseEventField,
    VideoHideTranscriptEventField,
    VideoShowTranscriptEventField,
)
from ralph.models.selector import selector

from ..browser import BaseBrowserModel

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class UILoadVideo(BaseBrowserModel):
    """Pydantic model for `load_video` statement.

    The browser emits this statement when the video is fully rendered and ready to
    play.

    Attributes:
        event (VideoBaseEventField): See VideoBaseEventField.
        event_type (str): Consists of the value `load_video`.
        name (str): Consists either of the value `load_video` or `edx.video.loaded`.
    """

    __selector__ = selector(event_source="browser", event_type="load_video")

    event: Union[
        Json[VideoBaseEventField],
        VideoBaseEventField,
    ]
    event_type: Literal["load_video"]
    name: Literal["load_video", "edx.video.loaded"]


class UIPlayVideo(BaseBrowserModel):
    """Pydantic model for `play_video` statement.

    The browser emits this statement when a user selects the video player's play
    control.

    Attributes:
        event (PlayVideoEventField): See PlayVideoEventField.
        event_type (str): Consists of the value `play_video`.
        name (str): Consists either of the value `play_video` or `edx.video.played`.
    """

    __selector__ = selector(event_source="browser", event_type="play_video")

    event: Union[
        Json[PlayVideoEventField],
        PlayVideoEventField,
    ]
    event_type: Literal["play_video"]
    name: Optional[Literal["play_video", "edx.video.played"]]


class UIPauseVideo(BaseBrowserModel):
    """Pydantic model for `pause_video` statement.

    The browser emits this statement when a user selects the video player's pause
    control.

    Attributes:
        event (PauseVideoEventField): See PauseVideoEventField.
        event_type (str): Consists of the value `pause_video`.
        name (str): Consists either of the value `pause_video` or `edx.video.paused`.
    """

    __selector__ = selector(event_source="browser", event_type="pause_video")

    event: Union[
        Json[PauseVideoEventField],
        PauseVideoEventField,
    ]
    event_type: Literal["pause_video"]
    name: Optional[Literal["pause_video", "edx.video.paused"]]


class UISeekVideo(BaseBrowserModel):
    """Pydantic model for `seek_video` statement.

    The browser emits this statement when a user selects a user interface control to go
    to a different point in the video file.

    Attributes:
        event (SeekVideoEventField): See SeekVideoEventField.
        event_type (str): Consists of the value `seek_video`.
        name (str): Consists either of the value `seek_video` or
            `edx.video.position.changed`.
    """

    __selector__ = selector(event_source="browser", event_type="seek_video")

    event: Union[
        Json[SeekVideoEventField],
        SeekVideoEventField,
    ]
    event_type: Literal["seek_video"]
    name: Optional[Literal["seek_video", "edx.video.position.changed"]]


class UIStopVideo(BaseBrowserModel):
    """Pydantic model for `stop_video` statement.

    The browser emits this statement when the video player reaches the end of the video
    file and play automatically stops.

    Attributes:
        event (StopVideoEventField): See StopVideoEventField.
        event_type (str): Consists of the value `stop_video`.
        name (str): Consists either of the value `stop_video` or `edx.video.stopped`.
    """

    __selector__ = selector(event_source="browser", event_type="stop_video")

    event: Union[
        Json[StopVideoEventField],
        StopVideoEventField,
    ]
    event_type: Literal["stop_video"]
    name: Optional[Literal["stop_video", "edx.video.stopped"]]


class UIHideTranscript(BaseBrowserModel):
    """Pydantic model for `hide_transcript` statement.

    The browser emits this statement when a user selects <kbd>CC</kbd> to suppress
    display of the video transcript.

    Attributes:
        event (VideoTranscriptEventField): See VideoTranscriptEventField.
        event_type (str): Consists of the value `hide_transcript`.
        name (str): Consists either of the value `hide_transcript` or
            `edx.video.transcript.hidden`.
    """

    __selector__ = selector(event_source="browser", event_type="hide_transcript")

    event: Union[
        Json[VideoHideTranscriptEventField],
        VideoHideTranscriptEventField,
    ]
    event_type: Literal["hide_transcript"]
    name: Literal["hide_transcript", "edx.video.transcript.hidden"]


class UIShowTranscript(BaseBrowserModel):
    """Pydantic model for `show_transcript` statement.

    The browser emits this statement when a user selects <kbd>CC</kbd> to display the
    video transcript.

    Attributes:
        event (VideoTranscriptEventField): See VideoTranscriptEventField.
        event_type (str): Consists of the value `show_transcript`.
        name (str): Consists either of the value `show_transcript` or
            `edx.video.transcript.shown`.
    """

    __selector__ = selector(event_source="browser", event_type="show_transcript")

    event: Union[
        Json[VideoShowTranscriptEventField],
        VideoShowTranscriptEventField,
    ]
    event_type: Literal["show_transcript"]
    name: Literal["show_transcript", "edx.video.transcript.shown"]


class UISpeedChangeVideo(BaseBrowserModel):
    """Pydantic model for `speed_change_video` statement.

    The browser emits this statement when a user selects a different playing speed for
    the video.

    Attributes:
        event (SpeedChangeVideoEventField): See SpeedChangeVideoEventField.
        event_type (str): Consists of the value `speed_change_video`.
    """

    __selector__ = selector(event_source="browser", event_type="speed_change_video")

    event: Union[
        Json[SpeedChangeVideoEventField],
        SpeedChangeVideoEventField,
    ]
    event_type: Literal["speed_change_video"]
    name: Optional[Literal["speed_change_video"]]


class UIVideoHideCCMenu(BaseBrowserModel):
    """Pydantic model for `video_hide_cc_menu` statement.

    The browser emits this statement when a user selects a language from the CC menu
    for a video that has transcripts in multiple languages

    Attributes:
        event (VideoBaseEventField): See VideoBaseEventField.
        event_type (str): Consists of the value `video_hide_cc_menu`.
    """

    __selector__ = selector(event_source="browser", event_type="video_hide_cc_menu")

    event: Union[
        Json[VideoBaseEventField],
        VideoBaseEventField,
    ]
    event_type: Literal["video_hide_cc_menu"]
    name: Optional[Literal["video_hide_cc_menu"]]


class UIVideoShowCCMenu(BaseBrowserModel):
    """Pydantic model for `video_show_cc_menu` statement.

    The browser emits this statement when a user selects CC for a video that has
    transcripts in multiple languages.

    Note: This statement is emitted in addition to the show_transcript event.

    Attributes:
        event (VideoBaseEventField): See VideoBaseEventField.
        event_type (str): Consists of the value `video_show_cc_menu`.
    """

    __selector__ = selector(event_source="browser", event_type="video_show_cc_menu")

    event: Union[
        Json[VideoBaseEventField],
        VideoBaseEventField,
    ]
    event_type: Literal["video_show_cc_menu"]
    name: Optional[Literal["video_show_cc_menu"]]



"""Navigation xAPI event definitions."""

from ...selector import selector
from ..base.statements import BaseXapiStatement
from ..concepts.activity_types.activity_streams_vocabulary import PageActivity
from ..concepts.verbs.scorm_profile import TerminatedVerb
from ..concepts.verbs.tincan_vocabulary import ViewedVerb


class PageViewed(BaseXapiStatement):
    """Pydantic model for page viewed statement.

    Example: John viewed the https://www.fun-mooc.fr/ page.

    Attributes:
       object (dict): See PageActivity.
       verb (dict): See ViewedVerb.
    """

    __selector__ = selector(
        object__definition__type="http://activitystrea.ms/schema/1.0/page",
        verb__id="http://id.tincanapi.com/verb/viewed",
    )

    object: PageActivity
    verb: ViewedVerb = ViewedVerb()


class PageTerminated(BaseXapiStatement):
    """Pydantic model for page terminated statement.

    Example: John terminated the https://www.fun-mooc.fr/ page.

    Attributes:
       object (dict): See PageActivity.
       verb (dict): See TerminatedVerb.
    """

    __selector__ = selector(
        object__definition__type="http://activitystrea.ms/schema/1.0/page",
        verb__id="http://adlnet.gov/expapi/verbs/terminated",
    )

    object: PageActivity
    verb: TerminatedVerb = TerminatedVerb()
